/*******************************************************************************
* 
*	A test MPI implementation for Windows NT:
* 
* 	WINNT_basic.c: 
* 	contains tentative versions of
* 	the basic comminications functions ...
*
* int MPID_ControMsgAvail (void)
*
* void MPID_RecvAnyControl (MPID_WINNT_PKT_T* pkt, int size, int* from)
*
* void MPID_SendControl (MPID_WINNT_PKT_T* pkt, int size, int dest)
*
* void MPID_RecvFromChannel (void* buff, int maxsize, int from)
* 
* void MPID_SendChannel (void* buff, int size, int dest)
*
* 	as well as ... (not yet)
*
*	void MPID_IRecvFromChanel (void* buff, int size, int partner, int id) 
*
*	void MPID_WRecvFromChanel (void* buff, int size, int partner, int id)
*
*	void MPID_RecvStatus (int id)
*
*	void MPID_ISendChanel (void* buff, int size, int partner, int id)
*
*	void MPID_WSendChanel (void* buff, int size, int partner, int id)
*
*	void MPID_TSendChanel (int id)
*
* 	which are optional.
*
* Programmer: Boris V. Protopopov. 
* Mississippi State University, 
* ICDCR Lab.
* 
* Last update: 
* 01.18.95.
*
********************************************************************************/                



#include <stdio.h>
#include <windows.h>
#include <windowsx.h>
#include <memory.h>
#include "packets.h"
#include "queue.h"
#include "queuedata.h"
#include "processinfo.h"
#include "infoprint.h"
#include <crtdbg.h>

static MPID_PKT_T *pPendingSendPacket = NULL;
static MPID_PKT_T *pPendingReceivePacket = NULL;

/* Function definitions */

int MPID_ControlMsgAvail (void)
/* The function retuens TRUE if there is any control packet for the process;
   otherwise it returns FALSE */
{
	if (IsEmptyQueue(RECEIVEQUEUE)) 
		return 0;
	else 
		return 1;
}
		
void MPID_RecvAnyControl (MPID_PKT_T *pkt, int size, int *from)
/*   
   !!! Note: the function blocks until something is received. 

*/
{
	DWORD dwStatus;
	BOOL fTruncate = FALSE;
	MPID_PKT_T *pTmpPkt;
	DWORD dwPacketSize;
	extern QUEUETYPE Queue[];


	_ASSERTE(pkt);
	_ASSERTE(0 <= size);
	_ASSERTE(from);

	dwStatus = WaitForSingleObject(Queue[RECEIVEQUEUE].hSomethingQueued,INFINITE);
	_ASSERTE(WAIT_FAILED != dwStatus);
	if (WAIT_FAILED == dwStatus)
		goto reporterror;

	Dequeue(RECEIVEQUEUE,&pTmpPkt,from,&dwStatus);
	_ASSERTE(ERROR_SUCCESS == dwStatus);
	if (ERROR_SUCCESS != dwStatus)
		goto reporterror;
	switch (pTmpPkt->head.mode)
	{
	case MPID_PKT_SHORT:
		dwPacketSize =    sizeof(MPID_PKT_HEAD_T)
						+ pTmpPkt->head.len;
		_ASSERTE((int)dwPacketSize <= size);
		if ((int)dwPacketSize > size)
		{
			pTmpPkt->head.len = size - sizeof(MPID_PKT_HEAD_T);
			dwPacketSize = size;
			fTruncate = TRUE;
		}
		_ASSERTE(0 <= (int)dwPacketSize);
		memcpy(pkt,pTmpPkt,dwPacketSize);
		free(pTmpPkt);
		break;
	case MPID_PKT_SHORT_SYNC:
		dwPacketSize =    sizeof(MPID_PKT_HEAD_T)
						+ sizeof(DECL_SYNC_ID_T)
						+ pTmpPkt->head.len;
		_ASSERTE((int)dwPacketSize <= size);
		if ((int)dwPacketSize > size)
		{
			pTmpPkt->head.len = size - sizeof(MPID_PKT_HEAD_T);
			dwPacketSize = size;
			fTruncate = TRUE;
		}
		_ASSERTE(0 <= (int)dwPacketSize);
		memcpy(pkt,pTmpPkt,dwPacketSize);
		free(pTmpPkt);
		break;
	case MPID_PKT_LONG:
		dwPacketSize = sizeof(MPID_PKT_LONG_T);
		_ASSERTE((int)dwPacketSize <= size);
		if ((int)dwPacketSize > size)
		{
			pTmpPkt->head.len = size - sizeof(MPID_PKT_HEAD_T);
			dwPacketSize = size;
			fTruncate = TRUE;
		}
		_ASSERTE(0 <= (int)dwPacketSize);
		memcpy(pkt,pTmpPkt,dwPacketSize);
		_ASSERTE(NULL == pPendingReceivePacket);
		pPendingReceivePacket = pTmpPkt;
		break;
	case MPID_PKT_LONG_SYNC:
		dwPacketSize = sizeof(MPID_PKT_LONG_SYNC_T);
		_ASSERTE((int)dwPacketSize <= size);
		if ((int)dwPacketSize > size)
		{
			pTmpPkt->head.len = size - sizeof(MPID_PKT_HEAD_T);
			dwPacketSize = size;
			fTruncate = TRUE;
		}
		_ASSERTE(0 <= (int)dwPacketSize);
		memcpy(pkt,pTmpPkt,dwPacketSize);
		_ASSERTE(NULL == pPendingReceivePacket);
		pPendingReceivePacket = pTmpPkt;
		break;
	case MPID_PKT_SYNC_ACK:
		dwPacketSize =    sizeof(MPID_PKT_SYNC_ACK_T);
		memcpy(pkt,pTmpPkt,dwPacketSize);
		free(pTmpPkt);
		break;
	default:
		/*
			Don't know what it is but there must be an
			internal error of some sort.
		*/
		_ASSERTE(FALSE);
		break;
	}
	//pkt->head.src &= 0x0000FFFF;
	_ASSERTE(!fTruncate);
cleanup:
	_ASSERTE(ERROR_SUCCESS == dwStatus);
	return;

reporterror:
	dwStatus = GetLastError();
	goto cleanup;
}



void MPID_SendControl (MPID_PKT_T *pkt, int size, int dest)
/* 
	The function sends a packet (pointed by 'pkt') of size 'size' to the destination
	'dest'. It is important that the function did not block (it will wait for the
	mutex guarding the buffer it is going to write into anyway - it is not considered
	blocking here (so far)). The function returns TRUE if the packet is successfully
	sent and FALSE otherwise because we cannot guarantee that the  dest buffer is free
	at the moment of the operation. 
*/
{
	DWORD dwStatus = ERROR_SUCCESS;
	DWORD dwPacketSize;
	BOOL fTruncate = FALSE;
	MPID_PKT_T *pTmpPkt = NULL;

	_ASSERTE(pkt);
	_ASSERTE(0 <= size);
	pTmpPkt = (MPID_PKT_T *)malloc(sizeof(MPID_PKT_T));
	_ASSERTE(pTmpPkt);
	if (NULL == pTmpPkt)
		goto reporterror;
	switch (pkt->head.mode)
	{
	case MPID_PKT_SHORT:
		dwPacketSize =	  sizeof(MPID_PKT_HEAD_T)
						+ pkt->head.len;
		_ASSERTE((int)dwPacketSize <= size);
		if ((int)dwPacketSize > size)
		{
			pTmpPkt->head.len = size - sizeof(MPID_PKT_HEAD_T);
			dwPacketSize = size;
			fTruncate = TRUE;
		}
		_ASSERTE(0 <= (int)dwPacketSize);
		memcpy(pTmpPkt,pkt,dwPacketSize);
		//pTmpPkt->head.src = MPID_MyWorldRank & 0x0000FFFF;
		//pTmpPkt->head.src |= (dest << 16);
		pTmpPkt->head.src = MPID_MyWorldRank;
		pTmpPkt->head.dest = dest;
		Enqueue(SENDQUEUE,pTmpPkt,MPID_MyWorldRank,&dwStatus);
		break;
	case MPID_PKT_SHORT_SYNC:
		dwPacketSize =	  sizeof(MPID_PKT_HEAD_T)
						+ sizeof(DECL_SYNC_ID_T)
						+ pkt->head.len;
		_ASSERTE((int)dwPacketSize <= size);
		if ((int)dwPacketSize > size)
		{
			pTmpPkt->head.len = size - sizeof(MPID_PKT_HEAD_T);
			dwPacketSize = size;
			fTruncate = TRUE;
		}
		_ASSERTE(0 <= (int)dwPacketSize);
		memcpy(pTmpPkt,pkt,dwPacketSize);
		//pTmpPkt->head.src = MPID_MyWorldRank & 0x0000FFFF;
		//pTmpPkt->head.src |= (dest << 16);
		pTmpPkt->head.dest = dest;
		pTmpPkt->head.src = MPID_MyWorldRank;
		Enqueue(SENDQUEUE,pTmpPkt,MPID_MyWorldRank,&dwStatus);
		break;
	case MPID_PKT_LONG:
		dwPacketSize =	  sizeof(MPID_PKT_LONG_T);
		_ASSERTE((int)dwPacketSize <= size);
		if ((int)dwPacketSize > size)
		{
			pTmpPkt->head.len = size - sizeof(MPID_PKT_HEAD_T);
			dwPacketSize = size;
			fTruncate = TRUE;
		}
		_ASSERTE(0 <= (int)dwPacketSize);
		memcpy(pTmpPkt,pkt,dwPacketSize);
		//pTmpPkt->head.src = MPID_MyWorldRank & 0x0000FFFF;
		//pTmpPkt->head.src |= (dest << 16);
		pTmpPkt->head.dest = dest;
		pTmpPkt->head.src = MPID_MyWorldRank;
		_ASSERTE(NULL == pPendingSendPacket);
		pPendingSendPacket = pTmpPkt;
		break;
	case MPID_PKT_LONG_SYNC:
		dwPacketSize =	  sizeof(MPID_PKT_LONG_SYNC_T);
		_ASSERTE((int)dwPacketSize <= size);
		if ((int)dwPacketSize > size)
		{
			pTmpPkt->head.len = size - sizeof(MPID_PKT_HEAD_T);
			dwPacketSize = size;
			fTruncate = TRUE;
		}
		_ASSERTE(0 <= (int)dwPacketSize);
		memcpy(pTmpPkt,pkt,dwPacketSize);
		//pTmpPkt->head.src = MPID_MyWorldRank & 0x0000FFFF;
		//pTmpPkt->head.src |= (dest << 16);
		pTmpPkt->head.dest = dest;
		pTmpPkt->head.src = MPID_MyWorldRank;
		_ASSERTE(NULL == pPendingSendPacket);
		pPendingSendPacket = pTmpPkt;
		break;
	case MPID_PKT_SYNC_ACK:
		dwPacketSize =	  sizeof(MPID_PKT_SYNC_ACK_T);
		memcpy(pTmpPkt,pkt,dwPacketSize);
		//pTmpPkt->head.src = MPID_MyWorldRank & 0x0000FFFF;
		//pTmpPkt->head.src |= (dest << 16);
		pTmpPkt->head.dest = dest;
		pTmpPkt->head.src = MPID_MyWorldRank;
		Enqueue(SENDQUEUE,pTmpPkt,MPID_MyWorldRank,&dwStatus);
		break;
	}
	_ASSERTE(ERROR_SUCCESS == dwStatus);
	if (ERROR_SUCCESS != dwStatus)
		goto reporterror;
	_ASSERTE(!fTruncate);

cleanup:
	return;
reporterror:
	dwStatus = GetLastError();
	goto cleanup;
}

void MPID_SendControlBlock (MPID_PKT_T *pkt, int size, int dest)
{
	_ASSERTE(pkt);
	_ASSERTE(0 <= size);
	MPID_SendControl (pkt, size, dest);
}

void MPID_RecvFromChannel (void *buff, int size, int from)
{
	DWORD dwStatus = ERROR_SUCCESS;
	DWORD dwBufferSize;
	BOOL fTruncate = FALSE;

	_ASSERTE(pPendingReceivePacket);
	_ASSERTE(buff);
	_ASSERTE(0 <= size);
	dwBufferSize = pPendingReceivePacket->head.len;
	if ((int)dwBufferSize > size)
	{
		pPendingReceivePacket->head.len = size - sizeof(MPID_PKT_HEAD_T);
		dwBufferSize = size;
		fTruncate = TRUE;
	}
	_ASSERTE(dwBufferSize);
	switch (pPendingReceivePacket->head.mode)
	{
	case MPID_PKT_LONG:
		_ASSERTE(pPendingReceivePacket->long_pkt.pBuffer);
		memcpy(buff,pPendingReceivePacket->long_pkt.pBuffer,dwBufferSize);
		free(pPendingReceivePacket->long_pkt.pBuffer);
		break;
	case MPID_PKT_LONG_SYNC:
		_ASSERTE(pPendingReceivePacket->long_sync_pkt.pBuffer);
		memcpy(buff,pPendingReceivePacket->long_sync_pkt.pBuffer,dwBufferSize);
		free(pPendingReceivePacket->long_sync_pkt.pBuffer);
		break;
	}
	free(pPendingReceivePacket);
	pPendingReceivePacket = NULL;
}


void MPID_SendChannel (void *buff, int size, int dest)
/*
	The function sends data (control is sent with the use of the functions
	described above).
*/
{
	DWORD dwStatus = ERROR_SUCCESS;
	BYTE *pBuffer = NULL;

	_ASSERTE(buff);
	_ASSERTE(0 <= size);
	pBuffer = (BYTE *)malloc(size);
	_ASSERTE(pBuffer);
	if (NULL == pBuffer)
		goto reporterror;
	memcpy(pBuffer,buff,size);
	_ASSERTE(pPendingSendPacket);
	switch (pPendingSendPacket->head.mode)
	{
	case MPID_PKT_LONG:
		pPendingSendPacket->long_pkt.pBuffer = pBuffer;
		break;
	case MPID_PKT_LONG_SYNC:
		pPendingSendPacket->long_sync_pkt.pBuffer = pBuffer;
		break;
	}
	Enqueue(SENDQUEUE,pPendingSendPacket,MPID_MyWorldRank,&dwStatus);
	if (dwStatus != ERROR_SUCCESS)
		goto reporterror;
	pPendingSendPacket = NULL;

cleanup:
	return;
reporterror:
	dwStatus = GetLastError();
	_ASSERTE(ERROR_SUCCESS != dwStatus);
	goto cleanup;
}
