#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <process.h>
#include <tchar.h>
#include <winsock.h>
#include "service.h"
#include "processinfo.h"
#include "mpiserve.h"
#include "MPIServePacket.h"
#include "MPIJobQueue.h"

static MPIJOBQUEUE JobQueue;

DWORD MPIJobQueueInitialize()
{
	DWORD dwStatus = ERROR_SUCCESS;

	InitializeCriticalSection(&JobQueue.cs);
	JobQueue.Qhead = NULL;
	JobQueue.fInitialized = TRUE;
	return dwStatus;
}

DWORD MPIJobQueueEnqueue(MPIJOBQUEUEINFO *pJobQInfo)
{
	DWORD dwStatus = (DWORD)~ERROR_SUCCESS;
	MPIJOBQUEUENODE *pJQN;

	if (JobQueue.fInitialized)
	{
		pJQN = (MPIJOBQUEUENODE*)malloc(sizeof(MPIJOBQUEUENODE));
		if (NULL == pJQN)
			goto reporterror;
		pJQN->pJob = pJobQInfo;
		pJQN->pNext = pJQN->pPrev = NULL;
		EnterCriticalSection(&JobQueue.cs);
		if (NULL == JobQueue.Qhead)
			/* there is nothing in the queue */
			JobQueue.Qhead = pJQN;
		else
		{
			pJQN->pNext = JobQueue.Qhead;
			pJQN->pNext->pPrev = pJQN;
			JobQueue.Qhead = pJQN;
		}
		LeaveCriticalSection(&JobQueue.cs);
		dwStatus = ERROR_SUCCESS;
	}
cleanup:
	return dwStatus;
reporterror:
	dwStatus = GetLastError();
	goto cleanup;
}

DWORD MPIJobQueueDequeue(char *szJobName,MPIJOBQUEUEINFO **ppJQI)
{
	DWORD dwStatus = (DWORD)~ERROR_SUCCESS;
	MPIJOBQUEUENODE *pLoop;

	*ppJQI = NULL;
	if (JobQueue.fInitialized)
	{
		EnterCriticalSection(&JobQueue.cs);
		for (pLoop = JobQueue.Qhead; pLoop; pLoop = pLoop->pNext)
		{
			if (!strcmp(szJobName,pLoop->pJob->szJobName))
			{
				*ppJQI = pLoop->pJob;
				pLoop->pJob = NULL;
				if (pLoop->pPrev)
					pLoop->pPrev->pNext = pLoop->pNext;
				else
					/* gotta be the head */
					JobQueue.Qhead = pLoop->pNext;
				if (pLoop->pNext)
					pLoop->pNext->pPrev = pLoop->pPrev;
				break;
			}
		}
		LeaveCriticalSection(&JobQueue.cs);
		if (pLoop)
			free(pLoop);
		dwStatus = ERROR_SUCCESS;
	}
	return dwStatus;
}