#include <crtdbg.h>
#include "windows.h"
#include "dm.h"
#include "packets.h"
#include "queue.h"
#include "queuedata.h"
#include "stdio.h"
#include "tchar.h"

#if !defined(SHMEM_AND_TCP) && !defined(SHMEM_ONLY)
#error "This module is for shared memory only!"
#endif


DWORD WINAPI SHMEMReceiveThread(LPVOID lpV)
{
	DWORD dwStatus = ERROR_SUCCESS;
	MPID_WINNT_SHMEM_BUF_T *pBuffers = 0;
	HANDLE *pEvents = 0;
	DWORD dwNumEvents;
	int i;
	DWORD dwShortPacketsReceived = 0;
	DWORD dwShortSyncPacketsReceived = 0;
	DWORD dwLongPacketsReceived = 0;
	DWORD dwLongSyncPacketsReceived = 0;
	DWORD dwSyncAckPacketsReceived = 0;
	PLONG_PKT_HOLDING_CELL pLongPackets = 0;

	extern int MPID_WorldSize;
	extern int MPID_MyWorldRank;
	extern HANDLE hEndProcessEvent;
	
	_ASSERTE(lpV);

	pBuffers = (MPID_WINNT_SHMEM_BUF_T *)lpV;
	dwNumEvents = MPID_WorldSize + 1;									   
	__MALLOC(pEvents, HANDLE, dwNumEvents);
	
	if (0 == pEvents)
	{
		dwStatus = GetLastError();
		_ASSERTE(FALSE);
	}
	
	pEvents[0] = hEndProcessEvent;
	for (i=0;i < MPID_WorldSize;i++)
		pEvents[i+1] = pBuffers[i].hBufferFilled;

	__MALLOC(pLongPackets, LONG_PKT_HOLDING_CELL, MPID_WorldSize);
	
	if (0 == pLongPackets)
	{
		dwStatus = GetLastError();
		_ASSERTE(FALSE);
	}
	for (i = 0;i < MPID_WorldSize; i++)
	{
		pLongPackets[i].pCopyTo = 0;
		pLongPackets[i].pPacket = 0;
		pLongPackets[i].dwNumBytes = 0;
	}
	while (TRUE)
	{
		MPID_WINNT_SHMEM_BUF_T *pShmemSourceBuffer;
		MPID_PKT_T *pTmpPkt;
		DWORD dwWhichEvent;
		DWORD dwWhichBuffer;
		DWORD dwBytesToSend;
		BOOL fStatus;
		LONG nErr;
		char *pDest;
		char *pSrc;

		/* listen for any buffer to fill */
		dwStatus = WaitForMultipleObjects(dwNumEvents,pEvents,FALSE,INFINITE);
		if (WAIT_FAILED == dwStatus)
		{
			dwStatus = GetLastError();
			_ASSERTE(FALSE);
		}
		dwWhichEvent = dwStatus - WAIT_OBJECT_0;
		if (0 == dwWhichEvent)
		{
			/* correct termination */
			goto cleanup;
		}
		dwWhichBuffer = dwWhichEvent - 1;
		pShmemSourceBuffer = &(pBuffers[dwWhichBuffer]);
		dwStatus = WaitForSingleObject(pShmemSourceBuffer->hMutex,INFINITE);
		if (WAIT_FAILED == dwStatus)
		{
			dwStatus = GetLastError();
			_ASSERTE(FALSE);
		}
		/* now I have the Mutex */
		

		switch (pShmemSourceBuffer->Packet.head.mode)
		{
		case MPID_PKT_SHORT:
		case MPID_PKT_SHORT_SYNC:
			/* go ahead and Malloc the buffer now */
			__MALLOC(pTmpPkt, MPID_PKT_T, 1);
			if (0 == pTmpPkt)
			{
				dwStatus = GetLastError();
				_ASSERTE(FALSE);
			}
			pTmpPkt->head = pShmemSourceBuffer->Packet.head;
			
			if(pTmpPkt->head.mode == MPID_PKT_SHORT) {
				pSrc = pShmemSourceBuffer->Packet.short_pkt.buffer;
				pDest = pTmpPkt->short_pkt.buffer;
				if(0 != pShmemSourceBuffer->Packet.head.len)
				{
					__MEMCPY(pDest,pSrc,pShmemSourceBuffer->Packet.head.len);
				}
			}
			else if(pTmpPkt->head.mode == MPID_PKT_SHORT_SYNC) {
				pTmpPkt->short_sync_pkt.sync_id = 
							pShmemSourceBuffer->Packet.short_sync_pkt.sync_id;
				pSrc = pShmemSourceBuffer->Packet.short_sync_pkt.buffer;
				pDest = pTmpPkt->short_sync_pkt.buffer;
				if(0 != pShmemSourceBuffer->Packet.head.len)
				{
					__MEMCPY(pDest,pSrc,pShmemSourceBuffer->Packet.head.len);
				}
			}	 

			fStatus = Enqueue(SHMEMRECEIVEQUEUE,pTmpPkt,pTmpPkt->head.from,&nErr);

			if (!fStatus)
			{
				_ASSERTE(FALSE);
			}
 			if (MPID_PKT_SHORT_SYNC == pShmemSourceBuffer->Packet.head.mode)
				dwShortSyncPacketsReceived++;
			else 
				dwShortPacketsReceived++;
			break;
		case MPID_PKT_LONG:
		case MPID_PKT_LONG_SYNC:
			__MALLOC(pTmpPkt, MPID_PKT_T, 1);
			if (0 == pTmpPkt)
			{
				dwStatus = GetLastError();
				_ASSERTE(FALSE);
			}
			__MEMCPY(pTmpPkt,&(pShmemSourceBuffer->Packet),sizeof(MPID_PKT_T));
			/* now, put it in a holding cell to be completed */
			__MALLOC(pTmpPkt->long_pkt.pBuffer, char, pTmpPkt->head.len);
			if (0 == pTmpPkt->long_pkt.pBuffer)
			{
				dwStatus = GetLastError();
				_ASSERTE(FALSE);
			}

			pLongPackets[dwWhichBuffer].pPacket = pTmpPkt;
			pLongPackets[dwWhichBuffer].pCopyTo = pTmpPkt->long_pkt.pBuffer;
			
			dwLongPacketsReceived++;

			break;
		case MPID_PKT_LONG_DATA:
			dwBytesToSend = 0;
			pTmpPkt = &(pShmemSourceBuffer->Packet);
			
			/* Check for exceeding length */
			if ((dwBytesToSend = pTmpPkt->head.len) > 
				(pLongPackets[dwWhichBuffer].pPacket->head.len - pLongPackets[dwWhichBuffer].dwNumBytes))
			{
				fprintf (stderr, "Trankated message from %d to %d.\n", 
								dwWhichBuffer,
								MPID_MyWorldRank);
				dwBytesToSend = pLongPackets[dwWhichBuffer].pPacket->head.len - 
									pLongPackets[dwWhichBuffer].dwNumBytes;
			}	
			/* Append curent data to the packet in the LONG_PKT_HOLDING_CELL */
			__MEMCPY(pLongPackets[dwWhichBuffer].pCopyTo,
					pTmpPkt->long_data_pkt.buffer,
					dwBytesToSend);
			pLongPackets[dwWhichBuffer].pCopyTo += dwBytesToSend;
			pLongPackets[dwWhichBuffer].dwNumBytes += dwBytesToSend;

			/* If all data is received, then enqueue */
			if (pLongPackets[dwWhichBuffer].dwNumBytes == (DWORD)pLongPackets[dwWhichBuffer].pPacket->head.len)
			{
				fStatus = Enqueue(SHMEMRECEIVEQUEUE,pLongPackets[dwWhichBuffer].pPacket,
								  pLongPackets[dwWhichBuffer].pPacket->head.from,&nErr);
				if (!fStatus)
				{
					_ASSERTE(FALSE);
				}
				pLongPackets[dwWhichBuffer].pCopyTo = 0;
				pLongPackets[dwWhichBuffer].pPacket = 0;
				pLongPackets[dwWhichBuffer].dwNumBytes = 0;
			}
			break;
		case MPID_PKT_SYNC_ACK:
			/* go ahead and Malloc the buffer now */

			__MALLOC(pTmpPkt, MPID_PKT_T, 1);
			if (0 == pTmpPkt)
			{
				dwStatus = GetLastError();
				_ASSERTE(FALSE);
			}
			pTmpPkt->sync_ack_pkt.mode = pShmemSourceBuffer->Packet.sync_ack_pkt.mode;
			pTmpPkt->sync_ack_pkt.sync_id = pShmemSourceBuffer->Packet.sync_ack_pkt.sync_id;
			fStatus = Enqueue(SHMEMRECEIVEQUEUE,pTmpPkt,pTmpPkt->head.from,&nErr);
			if (!fStatus)
			{
				_ASSERTE(FALSE);
			}
			dwSyncAckPacketsReceived++;
			break;
		}
		fStatus = SetEvent(pShmemSourceBuffer->hBufferEmptied);
		/* Ok, release it now */
		fStatus = ReleaseMutex(pShmemSourceBuffer->hMutex);
		if (!fStatus)
		{
			dwStatus = GetLastError();
			_ASSERTE(FALSE);
		}

	}

cleanup:
	if (pEvents)
		__FREE(pEvents);
	OutputDebugString("The SHMEMRECVThread terminates.");
	return dwStatus;
}
