/*******************************************************************************
* 
*	A test MPI implementation for Windows NT:
* 
* 	WINNT_basic.c: 
* 	contains tentative versions of
* 	the basic comminications functions ...
*
* int MPID_ControMsgAvail (void)
*
* void MPID_RecvAnyControl (MPID_WINNT_PKT_T* pkt, int size, int* from)
*
* void MPID_SendControl (MPID_WINNT_PKT_T* pkt, int size, int dest)
*
* void MPID_RecvFromChannel (void* buff, int maxsize, int from)
* 
* void MPID_SendChannel (void* buff, int size, int dest)
*
* 	as well as ... (not yet)
*
*	void MPID_IRecvFromChanel (void* buff, int size, int partner, int id) 
*
*	void MPID_WRecvFromChanel (void* buff, int size, int partner, int id)
*
*	void MPID_RecvStatus (int id)
*
*	void MPID_ISendChanel (void* buff, int size, int partner, int id)
*
*	void MPID_WSendChanel (void* buff, int size, int partner, int id)
*
*	void MPID_TSendChanel (int id)
*
* 	which are optional.
*
* Programmer: Boris V. Protopopov. 
* Mississippi State University, 
* ICDCR Lab.
* 
* Last update: 
* 01.18.95.
*
********************************************************************************/                



#include <stdio.h>
#include <windows.h>
#include <windowsx.h>
#include <memory.h>
#include "packets.h"
#include "queuedata.h"
#include "queue.h"

#include <crtdbg.h>

#if !defined(SHMEM_ONLY) && !defined(TCP_ONLY) && !defined(SHMEM_AND_TCP)
#error You must specify some sort of messaging method
#endif

static MPID_PKT_T* pCurrentRecvLongControl;
static MPID_PKT_T* pCurrentSendLongControl;

/* If several user threads are in progress, 
static char* pCurrentRecvLongData[MPID_WorldSize]; - next version
*/

/* Function definitions */

int MPID_ControlMsgAvail (void)
/* The function retuens TRUE if there is any control packet for the process;
   otherwise it returns FALSE */
{
#ifdef SHMEM_ONLY
	if (IsEmptyQueue(SHMEMRECEIVEQUEUE)) 
		return 0;
	else 
		return 1;
#endif
#ifdef TCP_ONLY
	if (IsEmptyQueue(TCPIPRECEIVEQUEUE)) 
		return 0;
	else 
		return 1;
#endif
#ifdef SHMEM_AND_TCP
	if (IsEmptyQueue(TCPIPRECEIVEQUEUE) && IsEmptyQueue(SHMEMRECEIVEQUEUE))
		return 0;
	else 
		return 1;
#endif
}
		
void MPID_RecvAnyControl (MPID_PKT_T* pkt, int size, int* from)
/*   
   !!! Note: the function blocks until something is received. 

*/
{
	extern QUEUETYPE Queue[NUMQUEUES];
	DWORD nStatus;
	MPID_PKT_T* pPacket;
	LONG actual_size;
	int HEAD_LEN = sizeof(MPID_PKT_HEAD_T);
	int HEAD_N_SID_LEN = sizeof(MPID_PKT_HEAD_T) + sizeof(LONG);

#ifdef SHMEM_AND_TCP
	DWORD i;
	HANDLE hArray[NUMQUEUES/2];
#endif

	_ASSERTE(_CrtIsValidPointer(pkt, sizeof(MPID_PKT_T), TRUE));

#ifdef SHMEM_ONLY
	nStatus = WaitForSingleObject(Queue[SHMEMRECEIVEQUEUE].hSomethingQueuedEvent, INFINITE);
#elif defined(TCP_ONLY)
	nStatus = WaitForSingleObject(Queue[TCPIPRECEIVEQUEUE].hSomethingQueuedEvent, INFINITE);
#elif defined(SHMEM_AND_TCP)
	hArray[0] = Queue[SHMEMRECEIVEQUEUE].hSomethingQueuedEvent;
	hArray[1] = Queue[TCPIPRECEIVEQUEUE].hSomethingQueuedEvent;
	nStatus = WaitForMultipleObjects(2,hArray,FALSE,INFINITE);
#endif

	if (WAIT_FAILED != nStatus)
	{
#ifdef SHMEM_ONLY
		Dequeue(SHMEMRECEIVEQUEUE, &pPacket, from, &nStatus);
#elif defined(TCP_ONLY)
		Dequeue(TCPIPRECEIVEQUEUE, &pPacket, from, &nStatus);
#elif defined(SHMEM_AND_TCP)
		if (0 == nStatus)
			Dequeue(SHMEMRECEIVEQUEUE, &pPacket, from, &nStatus);
		else if (1 == nStatus)
			Dequeue(TCPIPRECEIVEQUEUE, &pPacket, from, &nStatus);
#endif
		
		if(ERROR_SUCCESS != nStatus)
		{
			fprintf(stderr, "Error dequeueing a packet.\n");
			_ASSERTE(FALSE);
			MPID_WINNT_Abort();
		}
		switch (pPacket->head.mode)
		{
		case MPID_PKT_SHORT:
			if ((actual_size = pPacket->head.len + HEAD_LEN) > size)
			{
				actual_size = size;
				pkt->head.len = size - HEAD_LEN;
				fprintf (stderr, "Process %d: received trankated packet from %d.\n", 
									MPID_MyWorldRank, *from);
				_ASSERTE(FALSE);
			}

			_ASSERTE(_CrtIsValidPointer(pPacket, actual_size, TRUE));
	
			__MEMCPY(pkt, pPacket, actual_size);			
			__FREE(pPacket);
			break;
		case MPID_PKT_SHORT_SYNC:
			if ((actual_size = pPacket->head.len + HEAD_N_SID_LEN) > size)
			{
				actual_size = size;
				pkt->head.len = size - HEAD_N_SID_LEN;
				fprintf (stderr, "Process %d: received trankated packet from %d.\n", 
									MPID_MyWorldRank, *from);
				_ASSERTE(FALSE);
			}
			
			_ASSERTE(_CrtIsValidPointer(pPacket, actual_size, TRUE));
	
			__MEMCPY(pkt, pPacket, actual_size);
			__FREE(pPacket);
			break;
		case MPID_PKT_LONG:
			
			_ASSERTE(_CrtIsValidPointer(pPacket, sizeof(MPID_PKT_LONG_T), TRUE));
	
			__MEMCPY(pkt, pPacket, sizeof(MPID_PKT_LONG_T));
			pCurrentRecvLongControl = pPacket;
			/* do not free anything here */
			break;
		case MPID_PKT_LONG_SYNC:

			_ASSERTE(_CrtIsValidPointer(pPacket, sizeof(MPID_PKT_LONG_SYNC_T), TRUE));
			
			__MEMCPY(pkt, pPacket, sizeof(MPID_PKT_LONG_SYNC_T));
			pCurrentRecvLongControl = pPacket;
			/* do not free anything here */
			break;
		case MPID_PKT_SYNC_ACK:
			/* copy MODE fields */

			_ASSERTE(_CrtIsValidPointer(pPacket, sizeof(MPID_PKT_SYNC_ACK_T), TRUE));

			pkt->sync_ack_pkt.mode = pPacket->sync_ack_pkt.mode;
			//pkt->sync_ack_pkt.context_id = pPacket->sync_ack_pkt.context_id;
			//pkt->sync_ack_pkt.lrank = pPacket->sync_ack_pkt.lrank;
			
			/* copy sinc_id */
			pkt->sync_ack_pkt.sync_id = pPacket->sync_ack_pkt.sync_id;
			/* this is the whole MPID_PKT_SYNC_ACK */
			
			__FREE(pPacket);
			break;
		}

	}
	else 
	{
		fprintf(stderr, "Error waiting for hPacketAvailable event.\n");
		_ASSERTE(FALSE);
		MPID_WINNT_Abort();
	}

	return;
}



void MPID_SendControl (MPID_PKT_T* pkt, int size, int dest)
/* The function sends a packet (pointed by 'pkt') of size 'size' to the destination 'dest'. It is
   important that the function did not block (it will wait for the mutex guarding the buffer it is
   going to write into anyway - it is not considered blocking here (so far)). The function returns
   TRUE if the packet is successfully sent and FALSE otherwise because we cannot guarantee that the 
   dest buffer is free at the moment of the operation. 
*/
{
	LONG nStatus;
	MPID_PKT_T* pPacket;
	int actual_size;
	int HEAD_LEN = sizeof(MPID_PKT_HEAD_T);
	int HEAD_N_SID_LEN = sizeof(MPID_PKT_HEAD_T) + sizeof(LONG);

	/* Check partner first */

	CHECK_PARTNER (dest);

	__MALLOC(pPacket, MPID_PKT_T, 1);

	_ASSERTE(_CrtIsValidPointer(pkt, size, TRUE));

	_ASSERTE(_CrtIsValidPointer(pPacket, sizeof(MPID_PKT_T), TRUE));

	switch (pkt->head.mode)
	{
	case MPID_PKT_SHORT:

		_ASSERT(size >= HEAD_LEN);

		if ((actual_size = pkt->head.len + HEAD_LEN) > size)
		{
			actual_size = size;
			pPacket->head.len = size - HEAD_LEN;
			fprintf (stderr, "Process %d: sending trunkated packet to %d.\n", 
							MPID_MyWorldRank, dest);
			_ASSERTE(FALSE);
		}
		
		__MEMCPY(pPacket, pkt, actual_size);

#if defined(SHMEM_ONLY) || defined(SHMEM_AND_TCP)
		/*
		NOTE: In a system using both shared memory and TCP,
		the MPI process will ALWAYS send by shared memory.
		It is up to the SHMEMSendThread to figure out who
		the packet is delivered to.  Only in the case of
		TCP only will the MPI process handle its own TCP
		communications.  Of course, this means that the
		TCP proxy process will have to be careful if it
		uses this function.
		*/
		Enqueue(SHMEMSENDQUEUE, pPacket, dest, &nStatus);
#elif defined(TCP_ONLY)
		Enqueue(TCPIPSENDQUEUE, pPacket, dest, &nStatus);
#endif

		if(ERROR_SUCCESS != nStatus)
		{
			fprintf(stderr, "Error enqueueing a packet.\n");
			_ASSERTE(FALSE);
			MPID_WINNT_Abort();
		}
		break;
	case MPID_PKT_SHORT_SYNC:

		_ASSERT(size >= HEAD_LEN);

		if ((actual_size = pkt->head.len + HEAD_N_SID_LEN) > size)
		{
			actual_size = size;
			pPacket->head.len = size - HEAD_N_SID_LEN;
			fprintf (stderr, "Process %d: sending trankated packet to %d.\n", 
							MPID_MyWorldRank, dest);
			_ASSERTE(FALSE);
		}
		__MEMCPY(pPacket, pkt, actual_size);
#if defined(SHMEM_ONLY) || defined(SHMEM_AND_TCP)
		/*
		NOTE: In a system using both shared memory and TCP,
		the MPI process will ALWAYS send by shared memory.
		It is up to the SHMEMSendThread to figure out who
		the packet is delivered to.  Only in the case of
		TCP only will the MPI process handle its own TCP
		communications.  Of course, this means that the
		TCP proxy process will have to be careful if it
		uses this function.
		*/
		Enqueue(SHMEMSENDQUEUE, pPacket, dest, &nStatus);
#elif defined(TCP_ONLY)
		Enqueue(TCPIPSENDQUEUE, pPacket, dest, &nStatus);
#endif
		if(ERROR_SUCCESS != nStatus)
		{
			fprintf(stderr, "Error enqueueing a packet.\n");
			_ASSERTE(FALSE);
			MPID_WINNT_Abort();
		}
		break;
	case MPID_PKT_SYNC_ACK:

		_ASSERT(size >= (2 * sizeof(LONG)));

		/* copy MODE fields */
		pPacket->sync_ack_pkt.mode = pkt->sync_ack_pkt.mode;
		//pPacket->sync_ack_pkt.context_id = pkt->sync_ack_pkt.context_id;
		//pPacket->sync_ack_pkt.lrank = pkt->sync_ack_pkt.lrank;
		
		/* copy sinc_id */
		pPacket->sync_ack_pkt.sync_id = pkt->sync_ack_pkt.sync_id;
		/* this is the whole MPID_PKT_SYNC_ACK - two fields */

#if defined(SHMEM_ONLY) || defined(SHMEM_AND_TCP)
		/*
		NOTE: In a system using both shared memory and TCP,
		the MPI process will ALWAYS send by shared memory.
		It is up to the SHMEMSendThread to figure out who
		the packet is delivered to.  Only in the case of
		TCP only will the MPI process handle its own TCP
		communications.  Of course, this means that the
		TCP proxy process will have to be careful if it
		uses this function.
		*/
		Enqueue(SHMEMSENDQUEUE, pPacket, dest, &nStatus);
#elif defined(TCP_ONLY)
		Enqueue(TCPIPSENDQUEUE, pPacket, dest, &nStatus);
#endif
		if(ERROR_SUCCESS != nStatus)
		{
			fprintf(stderr, "Error enqueueing a packet.\n");
			_ASSERTE(FALSE);
			MPID_WINNT_Abort();
		}
		break;
	case MPID_PKT_LONG:

		_ASSERT(size == sizeof(MPID_PKT_LONG_T));

		__MEMCPY(pPacket, pkt, sizeof(MPID_PKT_LONG_T));

		if (0 != pCurrentSendLongControl) 
		{
			fprintf (stderr, "Overwriting some control info.\n");
			_ASSERT(FALSE);
		}
		pCurrentSendLongControl = pPacket;
		break;
	case MPID_PKT_LONG_SYNC:

		_ASSERT(size == sizeof(MPID_PKT_LONG_SYNC_T));

		__MEMCPY(pPacket, pkt, sizeof(MPID_PKT_LONG_SYNC_T));
		if (0 != pCurrentSendLongControl)
			fprintf (stderr, "Overwriting some control info.\n");
		pCurrentSendLongControl = pPacket;
		break;
	}
	
	return;
}
		
void MPID_SendControlBlock (MPID_PKT_T* pkt, int size, int dest)
{
	MPID_SendControl (pkt, size, dest);
}

void MPID_RecvFromChannel (void* buff, int size, int from)
{
	/* Check partner first */

	int actual_size = 0;

	CHECK_PARTNER(from);

	_ASSERTE(_CrtIsValidPointer(pCurrentRecvLongControl, 
								sizeof(MPID_PKT_T),
								TRUE));
	_ASSERTE(_CrtIsValidPointer(pCurrentRecvLongControl->long_pkt.pBuffer,
								pCurrentRecvLongControl->long_pkt.len,
								TRUE));

	_ASSERTE(_CrtIsValidPointer(buff, size, TRUE));

	if (0 != pCurrentRecvLongControl)
	{
		if ((actual_size = pCurrentRecvLongControl->long_pkt.len) <= size) 
		{
			__MEMCPY(buff, pCurrentRecvLongControl->long_pkt.pBuffer, actual_size);
		}
		else
		{
			fprintf (stderr, "Process %d received trunkated long data from %d.\n", 
							MPID_MyWorldRank, 
							from);
			__MEMCPY(buff, pCurrentRecvLongControl->long_pkt.pBuffer, size);
		}
		
		__FREE(pCurrentRecvLongControl->long_pkt.pBuffer);
		
		__FREE(pCurrentRecvLongControl);
		pCurrentRecvLongControl = (MPID_PKT_T*) 0;
	}
	else
	{
		fprintf (stderr, "Invalid long control pointer in function MPID_RecvFromChannel.\n\
						\r	Process %d was receiving long message from %d.\n", 
						MPID_MyWorldRank, 
						from);
		_ASSERTE(FALSE);
	}
	
	return;
}




void MPID_SendChannel (void* buff, int size, int dest)
/* The function sends data (control is sent with the use of the functions described above). */
{
	LONG nStatus;
	char* pBuffer;

	/* Check partner first */

	CHECK_PARTNER (dest);

	_ASSERTE(_CrtIsValidPointer(pCurrentSendLongControl, 
								sizeof(MPID_PKT_T),
								TRUE));
	_ASSERTE(_CrtIsValidPointer(buff, size, TRUE));


	__MALLOC(pBuffer, char, size);

	_ASSERTE(_CrtIsValidPointer(pBuffer, size, TRUE));

	__MEMCPY(pBuffer, buff, size);

	if (0 != pCurrentSendLongControl)
	{
		pCurrentSendLongControl->long_pkt.pBuffer = pBuffer;

		_ASSERTE(size == pCurrentSendLongControl->long_pkt.len);

		_ASSERTE(_CrtIsValidPointer(pCurrentSendLongControl->long_pkt.pBuffer,
									pCurrentSendLongControl->long_pkt.len,
									TRUE));

#if defined(SHMEM_ONLY) || defined(SHMEM_AND_TCP)
		/*
		NOTE: In a system using both shared memory and TCP,
		the MPI process will ALWAYS send by shared memory.
		It is up to the SHMEMSendThread to figure out who
		the packet is delivered to.  Only in the case of
		TCP only will the MPI process handle its own TCP
		communications.  Of course, this means that the
		TCP proxy process will have to be careful if it
		uses this function.
		*/
		Enqueue(SHMEMSENDQUEUE, pCurrentSendLongControl, dest, &nStatus);
#elif defined(TCP_ONLY)
		Enqueue(TCPIPSENDQUEUE, pCurrentSendLongControl, dest, &nStatus);
#endif
		if(ERROR_SUCCESS != nStatus)
		{
			fprintf(stderr, "Error enqueueing a packet.\n");
			_ASSERTE(FALSE);
			MPID_WINNT_Abort();
		}
		pCurrentSendLongControl = (MPID_PKT_T*) 0;
	}
	else
	{
		fprintf (stderr, "Invalid long control pointer in function MPID_SendChannel.\n\
						\r	Process %d was sending long message to %d.\n", 
						MPID_MyWorldRank, 
						dest);
	}
	
	return;
}	


